package com.grack.prologmachine;

import java.util.*;

/**
 * Implements the functionality for a unifiable predicate.  An atom is 
 * represented by a Predicate with arity 0 for simplicity.
 * @author Matthew Mastracci
 */
public class Predicate implements Unifiable {
    private PredicateSignature _signature;
    private LinkedList _arguments;

    /**
     * Constructs a new Predicate from a given name and arity.  Further 
     * arguments can be added with addArgument().
     * @param name The name of the predicate
     * @param arity The arity of the predicate
     */
    public Predicate( String name, int arity ) {
        _signature = new PredicateSignature( name, arity );
        _arguments = new LinkedList();
    }

    /**
     * Constructs a new Predicate from a given name and arity and a single
     * argument.
     * @param name The name of the predicate.
     * @param arity The arity of the predicate.
     * @param argument The argument to add to the predicate.
     */    
    public Predicate( String name, int arity, Unifiable argument ) {
        _signature = new PredicateSignature( name, arity );
        _arguments = new LinkedList();
        _arguments.addFirst( argument );
    }
    
    /** 
     * Constructs a new Predicate from a given name and arity and a list of
     * arguments.
     * @param name The name of the predicate.
     * @param arity The arity of the predicate.
     * @param arguments The arguments to add to the predicate.
     */
    public Predicate( String name, int arity, Unifiable[] arguments ) {
        _signature = new PredicateSignature( name, arity );
        _arguments = new LinkedList();
        for ( int i = 0; i < arguments.length; i++ ) {
            _arguments.addLast( arguments[i] );
        }
    }

    /**
     * Add a new argument to the predicate.
     * @param argument The argument to add.
     */
    public void addArgument( Unifiable argument ) {
        _arguments.addLast( argument );
    }

    /**
     * Replace an argument with another.
     * @param index The index of the argument.
     * @param argument The argument to replace at this index.
     */
    public void setArgument( int index, Unifiable argument ) {
        _arguments.set( index, argument );
    }

    /** 
     * Returns the number of arguments this predicate has.
     * @return The number of arguments this predicate has.
     */
    public int argumentCount() { 
        return _arguments.size();
    }

    /**
     * Returns whether this represents a valid predicate or not.
     * @return Whether this represents a valid predicate or not.
     */
    public boolean isValid() {
        return _signature.getArity() == _arguments.size();
    }

    /**
     * Computes the predicate signature for this predicate.
     * @return The predicate signature for this predicate.
     */
    public PredicateSignature getSignature() {
        return _signature;
    }

    /**
     * Returns the argument at the given index for this predicate.
     * @param index The index to retrieve the argument from.
     * @return The argument at the given index.
     */
    public Unifiable getArgument( int index ) {
        return (Unifiable)_arguments.get( index );
    }

    /**
     * Converts the predicate to a string representation.
     * @return The string representation.
     */
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append( _signature.getName() );
        if ( _arguments.size() > 0 ) {
            sb.append( "(" );
            for ( int i = 0; i < _arguments.size(); i++ ) {
                if ( i > 0 )
                    sb.append( "," );
                sb.append( _arguments.get( i ).toString() );
            }
            sb.append( ")" );
        }
        return sb.toString();
    }

    /**
     * Returns the name of the predicate.
     * @return The name of the predicate.
     */
    public String getName() {
        return _signature.getName();
    }

    /**
     * Returns the arity of the predicate.
     * @return The arity of the predicate.
     */
    public int getArity() {
        return _signature.getArity();
    }
}

