package com.grack.prologmachine;

/*
  A term is a general entity representing a parsed entity of Prolog code.

  The following are terms:

  [a,b,c] = .(a,.(b,.(c,[])))
  meows(X),has_tail(X) = ','(meows(X), has_tail(X))

  A term is composed of a tree of objects implementing Unifiable.
 */
public class Term implements Unifiable {
    public boolean isTrue() {
        return false;
    }
    
    public static boolean unify( Unifiable u1, Unifiable u2, MostGeneralUnifier mgu ) {
        /*
          Test to see if they are not unifiable first, by using the Herbrand
          algorithm.
        */

        // If u1 and u2 are integers, test to see if they are equal.
        if ( u1 instanceof IntegerNumber && u2 instanceof IntegerNumber ) {
            IntegerNumber i1 = (IntegerNumber)u1;
            IntegerNumber i2 = (IntegerNumber)u2;
            if ( i1.getValue() != i2.getValue() )
                return false;
        }

        // If u1 and u2 are both floats, test to see if they are equal.
        if ( u1 instanceof FloatNumber && u2 instanceof FloatNumber ) {
            FloatNumber f1 = (FloatNumber)u1;
            FloatNumber f2 = (FloatNumber)u2;
            if ( f1.getValue() != f2.getValue() )
                return false;
        }

        // If u1 and u2 are differing numeric types, they are not equal.
        if ( (u1 instanceof FloatNumber && u2 instanceof IntegerNumber) 
             || (u1 instanceof IntegerNumber && u2 instanceof FloatNumber ) ) {
            return false;
        }

        /* 
           If u1 and u2 are both predicates, test to see if they have the same
           the same name and arity.
        */
        if ( u1 instanceof Predicate && u2 instanceof Predicate ) {
            Predicate p1 = (Predicate)u1;
            Predicate p2 = (Predicate)u2;
            
            // if f1/n1 != f2/n2, they cannot unify.
            if ( !p1.getName().equals( p2.getName() ) || !(p1.getArity() == p2.getArity()) ) {
                return false;
            }

            /*
              Atomic terms unify with themselves (ie: cow=cow succeeds), 
              yielding no information.
            */
            if ( p1.getArity() == 0 )
                return true;

            /* 
               For two predicates, f(a1, a2, ... aN) = f(b1, b2, ... bN), 
               attempt to unify each seperate parameter 
               (ie: a1 = b1, a2 = b2, ... aN = bN). 
            */
            for ( int i = 0; i < p1.getArity(); i++ ) {
                // If we can't unify one of the parameters, we bail.
                if ( !unify( p1.getArgument( i ), p2.getArgument( i ), mgu ) ) 
                    return false;
                /*
                  Replace the instances of the variables bound in the rest of
                  the equations
                */
                mgu.applySubstitution( p1 );
                mgu.applySubstitution( p2 );
            }
        }

        if ( u1 instanceof Variable && u2 instanceof Variable ) {
            Variable v1 = (Variable)u1;            
            Variable v2 = (Variable)u2;

            // X unifies with X, but yields no information.
            if ( v1.getName().equals( v2.getName() ) )
                return true;
        }
        
        // If we have an equation of the form term = X, replace it with X = term
        if ( u2 instanceof Variable && !(u1 instanceof Variable) ) {
            Unifiable t = u1;
            u1 = u2;
            u2 = t;
        }

        /*
          If we have an equation of the form X = term, this unifies and 
          creates a substitution.
        */
        if ( u1 instanceof Variable ) {
            Variable v1 = (Variable)u1;
            // Create the substitution in the most general unifier.
            mgu.addSubstitution( v1, u2 );
            return true;
        }
        
        // Should never reach this step (in theory)
        return true;
    }
}
